(* metacirc-cont-ccc-mem.sml
 *
 * Meta-circular interpreter with explicit continuations, call/cc, and
 * mutable memory.
 *)
structure MetaCircContCCCMem = struct

    type var = OL.var

    datatype value = INTv of int
		   | BOOLv of bool
		   | FUNv of value * mem * cont -> value
		   | REFv of int

    withtype mem = { count: int, possess: int -> value }

    and cont = value * mem -> value

    fun vFUN  (FUNv f)  = f | vFUN _  = raise Fail "function expected"
    fun vBOOL (BOOLv b) = b | vBOOL _ = raise Fail "boolean expected"
    fun vINT  (INTv i)  = i | vINT _  = raise Fail "integer expected"

    fun nextref { count, possess } = REFv count
    fun augment ({ count, possess }, a) =
	{ count = count + 1,
	  possess = fn n => if n = count then a else possess n }
    fun update ({ count, possess }, REFv i, a) =
	{ count = count,
	  possess = fn n => if n = i then a else possess n }
      | update _ = raise Fail "reference expected"
    fun lookup ({ count, possess }, REFv i) = possess i
      | lookup _ = raise Fail "reference expected"

    type env = var -> value

    fun evcon (OL.INT i) = INTv i
      | evcon (OL.BOOL b) = BOOLv b

    fun eval (r, e, m, k) =
	case r of
	    OL.CONST c => k (evcon c, m)
	  | OL.VAR v => k (e v, m)
	  | OL.APPL { opr, opnd } =>
	      eval (opr, e, m, fn (f, m') =>
		eval (opnd, e, m', fn (a, m'') => vFUN f (a, m'', k)))
	  | OL.LAMBDA l => k (evlambda (l, e), m)
	  | OL.COND { prem, conc, altr } =>
	      eval (prem, e, m, fn (p, m') =>
		if vBOOL p then eval (conc, e, m', k)
		           else eval (altr, e, m', k))
	  | OL.LETREC { dvar, dexp, body } =>
	      let fun e' x = if x = dvar then evlambda (dexp, e') else e x
	      in eval (body, e', m, k)
	      end

    and evlambda ({ fp, body }, e) =
	FUNv (fn (a, m, k) => eval (body, ext (fp, a, e), m, k))

    and ext (z, a, e) x = if x = z then a else e x

    fun equal (INTv i, INTv j) = BOOLv (i = j)
      | equal (BOOLv x, BOOLv y) = BOOLv (x = y)
      | equal _ = raise Fail "illegal argument to equal"

    fun initenv "succ" = FUNv (fn (x, m, k) => k (INTv (vINT x + 1), m))
      | initenv "equal" =
	  FUNv (fn (x, m, k) =>
		   k (FUNv (fn (y, m, k) => k (equal (x, y), m)), m))
      | initenv "callcc" =
	  FUNv (fn (x, m, k) => vFUN x (FUNv (fn (x, m, _) => k (x, m)), m, k))
      | initenv "ref" =
	  FUNv (fn (a, m, k) => k (nextref m, augment (m, a)))
      | initenv "set" =
	  FUNv (fn (r, m, k) =>
		   k (FUNv (fn (a, m', k') => k' (a, update (m', r, a))), m))
      | initenv "val" =
	  FUNv (fn (r, m, k) => k (lookup (m, r), m))
      | initenv x = raise Fail ("unbound variable " ^ x)

    val initmem = { count = 0, possess = fn _ => INTv 0 }

    fun interpret r = eval (r, initenv, initmem, fn (a, _) => a)
end
