(* metacirc-cont-defunct-esc.sml
 *
 * Meta-circular intepreter with explicit continuations, defunctionalized,
 * "escape" (aka. "let/cc") added.
 *)
structure MetaCircContDefunctEsc = struct

    structure OL = OLEsc

    type var = OL.var

    datatype value = INTv of int
		   | BOOLv of bool
		   | FUNv of funval

    and funval = CLOSR of { lam: OL.lambda, en: env }
               | SC
	       | EQ1
	       | EQ2 of value
	       | ESC of cont

    and env = INIT
            | SIMP of { bvar: var, bval: value, old: env }
	    | REC of { dvar: var, dexp: OL.lambda, old: env }

    and cont = FIN
             | EVOPN of { opnd: OL.exp, en: env, next: cont }
	     | APFUN of { func: value, next: cont }
	     | BRANCH of { conc: OL.exp, altr: OL.exp, en: env, next: cont }

    fun vFUN  (FUNv f) = f  | vFUN _  = raise Fail "function expected"
    fun vBOOL (BOOLv b) = b | vBOOL _ = raise Fail "boolean expected"
    fun vINT  (INTv i) = i  | vINT _  = raise Fail "integer expected"

    fun evcon (OL.INT i) = INTv i
      | evcon (OL.BOOL b) = BOOLv b

    fun get (INIT, "succ") = FUNv SC
      | get (INIT, "equal") = FUNv EQ1
      | get (INIT, x) = raise Fail ("unbound variable: " ^ x)
      | get (SIMP { bvar, bval, old }, x) =
	  if bvar = x then bval else get (old, x)
      | get (e as REC { dvar, dexp, old }, x) =
	  if x = dvar then FUNv (CLOSR { lam = dexp, en = e })
	  else get (old, x)

    fun eval (r, e, k) =
	case r of
	    OL.CONST c => cont (k, evcon c)
	  | OL.VAR v => cont (k, get (e, v))
	  | OL.APPL { opr, opnd } =>
	      eval (opr, e, EVOPN { opnd = opnd, en = e, next = k })
	  | OL.LAMBDA l => cont (k, FUNv (CLOSR { lam = l, en = e }))
	  | OL.COND { prem, conc, altr } =>
	      eval (prem, e,
		    BRANCH { conc = conc, altr = altr, en = e, next = k })
	  | OL.LETREC { dvar, dexp, body } =>
	      eval (body, REC { dvar = dvar, dexp = dexp, old = e }, k)
	  | OL.ESCAPE { evar, body } =>
	     eval (body, SIMP { bvar = evar, bval = FUNv (ESC k), old = e }, k)

    and apply (CLOSR { lam = { fp, body }, en }, a, k) =
	  eval (body, SIMP { bvar = fp, bval = a, old = en }, k)
      | apply (SC, a, k) = cont (k, INTv (vINT a + 1))
      | apply (EQ1, a, k) = cont (k, FUNv (EQ2 a))
      | apply (EQ2 (INTv i), INTv j, k) = cont (k, BOOLv (i = j))
      | apply (EQ2 (BOOLv x), BOOLv y, k) = cont (k, BOOLv (x = y))
      | apply (EQ2 _, _, _) = raise Fail "illegal arguments to EQUAL"
      | apply (ESC k, a, _) = cont (k, a)

    and cont (FIN, a) = a
      | cont (EVOPN { opnd, en, next }, a) =
	  eval (opnd, en, APFUN { func = a, next = next })
      | cont (APFUN { func, next }, a) = apply (vFUN func, a, next)
      | cont (BRANCH { conc, altr, en, next }, a) =
	  if vBOOL a then eval (conc, en, next) else eval (conc, en, next)

    fun interpret r = eval (r, INIT, FIN)
end
