(* metacirc-defunct-all-cont.sml
 *
 * Introduced explicit continuations into completely defunctionalized
 * interpreter. (Continuations are functions, though, so the result is
 * not first-order anymore.)
 *)
structure MetaCircDefunctAllCont = struct

    type var = OL.var

    datatype value = INTv of int
		   | BOOLv of bool
		   | FUNv of funval

    and funval = CLOSR of { lam: OL.lambda, en: env }
               | SC
	       | EQ1
	       | EQ2 of value

    and env = INIT
            | SIMP of { bvar: var, bval: value, old: env }
	    | REC of { dvar: var, dexp: OL.lambda, old: env }

    fun vFUN  (FUNv f) = f  | vFUN _  = raise Fail "function expected"
    fun vBOOL (BOOLv b) = b | vBOOL _ = raise Fail "boolean expected"
    fun vINT  (INTv i) = i  | vINT _  = raise Fail "integer expected"

    fun evcon (OL.INT i) = INTv i
      | evcon (OL.BOOL b) = BOOLv b

    fun get (INIT, "succ") = FUNv SC
      | get (INIT, "equal") = FUNv EQ1
      | get (INIT, x) = raise Fail ("unbound variable: " ^ x)
      | get (SIMP { bvar, bval, old }, x) =
	  if bvar = x then bval else get (old, x)
      | get (e as REC { dvar, dexp, old }, x) =
	  if x = dvar then FUNv (CLOSR { lam = dexp, en = e })
	  else get (old, x)

    fun apply (CLOSR { lam = { fp, body }, en }, a, k) =
	  eval (body, SIMP { bvar = fp, bval = a, old = en }, k)
      | apply (SC, a, k) = k (INTv (vINT a + 1))
      | apply (EQ1, a, k) = k (FUNv (EQ2 a))
      | apply (EQ2 (INTv i), INTv j, k) = k (BOOLv (i = j))
      | apply (EQ2 (BOOLv x), BOOLv y, k) = k (BOOLv (x = y))
      | apply (EQ2 _, _, _) = raise Fail "illegal arguments to EQUAL"

    and eval (r, e, k) =
	case r of
	    OL.CONST c => k (evcon c)
	  | OL.VAR v => k (get (e, v))
	  | OL.APPL { opr, opnd } =>
	      eval (opr, e, fn f =>
		eval (opnd, e, fn a =>
		  apply (vFUN f, a, k)))
	  | OL.LAMBDA l => k (FUNv (CLOSR { lam = l, en = e }))
	  | OL.COND { prem, conc, altr } =>
	      eval (prem, e, fn p =>
		if vBOOL p then eval (conc, e, k) else eval (altr, e, k))
	  | OL.LETREC { dvar, dexp, body } =>
	      eval (body, REC { dvar = dvar, dexp = dexp, old = e }, k)

    fun interpret r = eval (r, INIT, fn a => a)
end
